import javax.swing.*;
import java.awt.*;
import java.awt.event.*;

/**
 * Graphical representation of a MIDI note.
 * 
 * @author  Andersson, Andreen, Lanneskog, Pehrson
 * @version 1
 */
 
public class MooNoteElement extends JPanel implements Comparable{

	private MooTrackView mtv;
	private MooNote note;
	private JPopupMenu popup;
	private JMenu popupTranspUp, popupTranspDown;
	private JMenuItem popupRemove, popupProp;
	private JMenuItem[] popupTranspUpItems, popupTranspDownItems;
	private Rectangle pitchRect, veloRect;
	private String notePitch, noteVelocity;
	private boolean selected = false, leftMouseButtonPressed = false, mouseIn = false;
	public Color textColor;
	public static final Color bgColor = new Color(160, 218, 255);
	public static final Color invBgColor = new Color(96, 38, 0);

	/** 
	 * Creates a new note element.
	 * @param parent	The MooTrackView that this element will be painted on.
	 * @param mn		the note that will be graphically represented
	 */
	public MooNoteElement (MooTrackView parent, MooNote mn) {
		mtv = parent;
		note = mn;
		calculateString();
		addMouseListener(new MAdapter());
		setBorder(BorderFactory.createLineBorder(Color.black));
		setBackground(bgColor);
		textColor = Color.black;

		// Defines coordinates.
		pitchRect = new Rectangle(0, 0, 15, 10);
		veloRect = new Rectangle(20, 0, 40, 10);

		// Creates pop-up menu.
		popup = new JPopupMenu();
		popupProp = addMenuItem(popup, "Preferences...");
		popupRemove = addMenuItem(popup, "Remove");
		popupTranspUpItems = new JMenuItem[12];
		popupTranspDownItems = new JMenuItem[12];
		popupTranspUp = createTransposeMenu(popup, popupTranspUpItems, "note up");
		popupTranspDown = createTransposeMenu(popup, popupTranspDownItems, "note down");
	}

	/** 
	 * Returns the note of this element.
	 * @return the note
	 */
	public MooNote getNote() {
		return note;
	}

	/** 
	 * Compares the note of this element to that of another note.
	 * @return a negative integer, zero, or a positive integer as this object is less than, equal to, or greater than the specified object
	 */
	public int compareTo(Object o) {
		return note.compareTo(((MooNoteElement)o).getNote());
	}

	/** 
	 * Selects the current NoteElement.
	 */
	public void select() {
		selected = true;
		mtv.selectNote(this);
		setBackground(invBgColor);
		textColor = Color.white;
		repaint();
	}

	/** 
	 * Deselects the current NoteElement.
	 */
	public void deselect() {
		selected = false;
		// mtv.deselectNote(this);
		setBackground(bgColor);
		textColor = Color.black;
		repaint();
	}

	/** 
	 * Transposes the current note element the given number of halftones.
	 * @param halftones	the number of halftones to transpose - positive for up, negative for down
	 */
	public void transpose(int halftones) {
		note.transpose(halftones);
		update();
	}

	/**
	 * Draws the string that shows the note's properties.
	 * @param g	The Graphics object used to draw the strings.
	 */
	public void paintComponent(Graphics g)
	{
		super.paintComponent(g);
		if (!(g instanceof Graphics2D)) return;
		Graphics2D g2 = (Graphics2D)g;
		g2.setColor(textColor);
		g2.setFont(new Font("Helvetica", Font.PLAIN, 8));
	/*
		switch(columns) {
			case 0:
			case 1:
			...
		}
	*/
		
		g2.drawString(notePitch, 1, 8);
		g2.drawString("" + noteVelocity, 21, 8);
	}

	/**
	 * Calculate what the string that shows the note properties should look like.
	 */
	protected void calculateString(){

		noteVelocity = ""; 
		notePitch = "";
		if(note == null) return;

		int pitch = note.getPitch();
		switch (pitch % 12) {
			case  0: notePitch = "C";  break;
			case  1: notePitch = "C#"; break;
			case  2: notePitch = "D";  break;
			case  3: notePitch = "D#"; break;
			case  4: notePitch = "E";  break;
			case  5: notePitch = "F";  break;
			case  6: notePitch = "F#"; break;
			case  7: notePitch = "G";  break;
			case  8: notePitch = "G#"; break;
			case  9: notePitch = "A";  break;
			case 10: notePitch = "A#"; break;
			case 11: notePitch = "B";  break;
		}
		notePitch += pitch / 12 - 1;
		noteVelocity = ""+note.getVelocity();
	}


	/**
	 * Asks the MooTrackView that the note element is painted on to remove this element and the note.
	 */
	protected void remove(){
		mtv.removeNote(this);
	}

	/**
	 * Updates the graphical content of the element and repaints it.
	 */
	public void update() {
		calculateString();
		repaint();
	}

	/**
	 * Layout this changed elemnt.
	 */
	protected void newLayout(){
		mtv.layoutElement(this,true);
	}

	/**
	 * Adds a menu item with the given command to the given popup menu.
	 */
	private JMenuItem addMenuItem(JPopupMenu menu, String command) {
		JMenuItem item = new JMenuItem(command);
		item.addActionListener(new PopupListener());
		menu.add(item);
		return item;
	}

	/**
	 * Adds a menu item with the given command to the given menu.
	 */
	private JMenuItem addMenuItem(JMenu menu, String command) {
		JMenuItem item = new JMenuItem(command);
		item.addActionListener(new PopupListener());
		menu.add(item);
		return item;
	}

	/**
	 * Creates a transpose sub menu with the given title in the given popup menu,
	 * inserting the items into the given array.
	 */
	private JMenu createTransposeMenu(JPopupMenu menu, JMenuItem[] items, String title) {
		JMenu trans = new JMenu("Transpose " + title);
		menu.add(trans);
		items[0] = addMenuItem(trans, "One octave");
		for (int i = 1; i < 12; i++) {
			items[i] = addMenuItem(trans, (i) + " halftones");
		}
		return trans;
	}

	/**
	 * Listener that checks the mouse actions on this element.
	 */
	class MAdapter extends MouseAdapter {
	
		/**
		 * Plays the note on double-click.
		 */
		public void mouseClicked(MouseEvent e) {
			if (SwingUtilities.isLeftMouseButton(e) && e.getClickCount() == 2)
				Moosique.getReceiver().send(note.getMessage(), -1);
		}

		/**
		 * Selects the note if mouse entered with the left mouse button pressed.
		 */
		public void mouseEntered(MouseEvent e) {
			mouseIn = true;
			if (mtv.isLeftMouseButtonPressed()) {
				select();
			}
		}

		/**
		 * Registers mouse exited.
		 */
		public void mouseExited(MouseEvent e) {
			mouseIn = false;
			Moosique.getReceiver().send(note.getNoteOffEvent().getMessage(), -1);
		}

		/**
		 * Checks if the mouse is pressed.
		 * Increases the pitch or velocity if the right mouse button is pressed while holding CTRL.
		 * Decreases the pitch or velocity if the left mouse button is pressed while holding CTRL.
		 * @param e the event recieved.
		 */
		public void mousePressed(MouseEvent e) {
			if (e.isControlDown()) {
				if (pitchRect.contains(e.getPoint())) {
					if (SwingUtilities.isRightMouseButton(e)) {
						note.transpose(1);
					} else if (SwingUtilities.isLeftMouseButton(e)) {
						note.transpose(-1);
					}
					Moosique.setEdited();
					calculateString();
					repaint();
				} else if (veloRect.contains(e.getPoint())) {
					if (SwingUtilities.isRightMouseButton(e)) {
						note.setVelocity(note.getVelocity() + 1);
					} else if (SwingUtilities.isLeftMouseButton(e)) {
						note.setVelocity(note.getVelocity() - 1);
					}
					Moosique.setEdited();
					calculateString();
					repaint();
				}
			} else {
				select();
				maybeShowPopup(e);
			}
		}

		public void mouseReleased(MouseEvent e) {
			if (!maybeShowPopup(e) && !mouseIn) {
				int y = e.getY();
				if (y < 0) mtv.maybeMoveSelectedNotes((int)Math.floor((double)y / MooTrackView.NOTE_HEIGHT) * MooTrackView.NOTE_HEIGHT);
				if (y > getHeight()) mtv.maybeMoveSelectedNotes((int)Math.ceil(((double)y - getHeight()) / MooTrackView.NOTE_HEIGHT) * MooTrackView.NOTE_HEIGHT);
				
			}
		}

		/**
		 * Shows the menu if an OS-specific popup-trigger was activated.
		 */
		private boolean maybeShowPopup(MouseEvent e) {
			if (!e.isPopupTrigger()) return false;
			if (!e.isControlDown()) {
				if (!selected || mtv.isTheOnlySelected((MooNoteElement)e.getComponent())) popup.show(e.getComponent(), e.getX(), e.getY());
				else mtv.showSelectionPopup(e.getComponent(), e.getX(), e.getY());
			}
			return true;
		}
	}

	/**
	 * Takes the appropriate action when a user selects an item on the popup menu.
	 */
	class PopupListener implements ActionListener {
		public void actionPerformed(ActionEvent e) {
			Object source = e.getSource();
			if (source == popupProp) {
				System.out.println("Duration: " + note.getDuration());
				new MooDialog(note);
				System.out.println("Duration: " + note.getDuration());
				newLayout();
				update();
			} else if (source == popupRemove) {
				remove();
			} else if (source == popupTranspUpItems[0]) {
				transpose(12);
			} else if (source == popupTranspDownItems[0]) {
				transpose(-12);
			} else {
				for (int i = 1; i < 12; i++) {
					if (source == popupTranspUpItems[i]) transpose(i);
					else if (source == popupTranspDownItems[i]) transpose(-i);
				}
			}
		}
	}
}
